/*
 *  AFS for BeOS - The native AtheOS file-system for BeOS
 *  Copyright (C) 2001 Ronny Schoebel
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <drivers/Drivers.h>
#include <support/SupportDefs.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <malloc.h>
#include <app/Message.h>
#include <interface/Screen.h>
#include <interface/Rect.h>
#include "fsproto.h"
#include "drive_setup.h"
#include "afsbuild.h"
#include "afs.h"
#include "InitializeWindow.h"

#define DPRINTF printf

static bool afs_validate_superblock( AfsSuperBlock_s* psSuperBlock, bool bQuiet )
{
    bool	bResult = true;
	
    if ( psSuperBlock->as_nMagic1 != SUPER_BLOCK_MAGIC1 )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Invalid super block magic1 = %lx\n", psSuperBlock->as_nMagic1 );
		bResult = false;
    }
    if ( psSuperBlock->as_nMagic2 != ( int32 )SUPER_BLOCK_MAGIC2 )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Invalid super block magic2 = %lx\n", psSuperBlock->as_nMagic2 );
		bResult = false;
    }
    if ( psSuperBlock->as_nMagic3 != SUPER_BLOCK_MAGIC3 )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Invalid super block magic3 = %lx\n", psSuperBlock->as_nMagic3 );
		bResult = false;
    }
    if ( psSuperBlock->as_nByteOrder != BO_LITTLE_ENDIAN && psSuperBlock->as_nByteOrder != BO_BIG_ENDIAN )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Bad value in super block byte-order field %ld\n", psSuperBlock->as_nByteOrder );
		bResult = false;
    }
    if ( ( uint32 )( 1 << psSuperBlock->as_nBlockShift ) != psSuperBlock->as_nBlockSize )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Inconsistent values in block-shift, and block-size values of superblock (sh=%ld/si=%ld)\n",
			psSuperBlock->as_nBlockShift, psSuperBlock->as_nBlockSize );
		bResult = false;
    }
    if ( psSuperBlock->as_nUsedBlocks > psSuperBlock->as_nNumBlocks )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Super block has more free blocks than actual block (%Ld - %Ld)\n",
			psSuperBlock->as_nUsedBlocks, psSuperBlock->as_nNumBlocks );
		bResult = false;
    }
    if ( ( uint32 )psSuperBlock->as_nInodeSize != psSuperBlock->as_nBlockSize )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Inconsistency between block-size (%ld) and inode-size (%ld)\n",
			psSuperBlock->as_nBlockSize, psSuperBlock->as_nInodeSize );
		bResult = false;
    }
    if ( 0 == psSuperBlock->as_nBlockSize ||
		( int32 )( ( 1 << psSuperBlock->as_nAllocGrpShift) / psSuperBlock->as_nBlockSize ) != psSuperBlock->as_nBlockPerGroup )
	{
		if ( bQuiet ) return( false );
		DPRINTF( "Inconsistency between blocks per ag (%ld) and ag shift count (%ld)\n",
			psSuperBlock->as_nBlockPerGroup, psSuperBlock->as_nAllocGrpShift );
		bResult = false;
    }
    if ( psSuperBlock->as_nAllocGroupCount * psSuperBlock->as_nBlockPerGroup < psSuperBlock->as_nNumBlocks )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Alloc group count (%ld) does not cower entire disk (%Ld)\n",
			psSuperBlock->as_nAllocGroupCount, psSuperBlock->as_nNumBlocks );
		bResult = false;
    }
    if ( psSuperBlock->as_sLogBlock.len != psSuperBlock->as_nLogSize )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Mismatch between length in journal run %d an the length recorded in the superblock %d\n",
			psSuperBlock->as_sLogBlock.len, psSuperBlock->as_nLogSize );
		bResult = false;
    }
    if ( psSuperBlock->as_nValidLogBlocks > psSuperBlock->as_nLogSize )
    {
		if ( bQuiet ) return( false );
		DPRINTF( "Superblock has more valid journal blocks (%d) than the total journal size (%d)\n",
			psSuperBlock->as_nValidLogBlocks, psSuperBlock->as_nLogSize );
		bResult = false;
    }
    return( bResult );
}

static int afs_probe( int32 nDev, uint64 nBase, fs_info* psInfo )
{
	AfsSuperBlock_s*	psSuperBlock;
	int					nError = 0;
	
	
	psSuperBlock = ( AfsSuperBlock_s* )malloc( AFS_SUPERBLOCK_SIZE );
	
	if ( psSuperBlock == NULL )
	{
		DPRINTF( "Error : afs_probe() Failed to alloc memory for super block\n" );
		nError = ENOMEM;
		goto error1;
	}
	
	if ( read_pos( nDev, nBase + 1024, psSuperBlock, AFS_SUPERBLOCK_SIZE ) != AFS_SUPERBLOCK_SIZE )
	{
		nError = EIO;
		goto error2;
	}

	if ( afs_validate_superblock( psSuperBlock, true ) == false )
	{
		if ( read_pos( nDev, nBase + 0, psSuperBlock, AFS_SUPERBLOCK_SIZE ) != AFS_SUPERBLOCK_SIZE )
		{
			DPRINTF( "afs_probe() Failed to read super block\n" );
			nError = EIO;
			goto error2;
		}
	}
	if ( afs_validate_superblock( psSuperBlock, true ) == false )
	{
		nError = EINVAL;
		goto error2;
	}
	psInfo->dev				= -1;
	psInfo->root			= afs_run_to_num( psSuperBlock, &psSuperBlock->as_sRootDir );
	psInfo->flags			= B_FS_IS_PERSISTENT | B_FS_HAS_MIME | B_FS_HAS_ATTR;
	psInfo->block_size		= psSuperBlock->as_nBlockSize;
	psInfo->io_size			= 65536;
	psInfo->total_blocks	= psSuperBlock->as_nNumBlocks;	
	psInfo->free_blocks		= psSuperBlock->as_nNumBlocks - psSuperBlock->as_nUsedBlocks;
	psInfo->total_nodes		= -1;
	psInfo->free_nodes		= -1;
	strncpy( psInfo->volume_name, psSuperBlock->as_zName, B_FILE_NAME_LENGTH );
	psInfo->volume_name[B_FILE_NAME_LENGTH - 1] = '\0';
//	strncpy( psInfo->device_name, pzDevPath, sizeof( psInfo->device_name ) );
	psInfo->device_name[sizeof( psInfo->device_name ) - 1] = '\0';
	strncpy( psInfo->fsh_name, "AtheOS-FS", B_OS_NAME_LENGTH );
	psInfo->fsh_name[B_OS_NAME_LENGTH - 1] = '\0'; 
	nError = B_OK;
	printf("AtheOS-FS volume name: %s (%s)\n", psSuperBlock->as_zName, psInfo->volume_name);
error2:
	free( psSuperBlock );
error1:
	return( nError );
}

void ds_fs_flags( drive_setup_fs_flags *psFlags )
{
	psFlags->can_initialize = true;
	psFlags->has_options = true;
}

bool ds_fs_id(partition_data *psPart, int32 nDev, uint64 nOffset, int32 nBlockSize)
{
	bool result = false;
	int nError;
	fs_info sFSInfo;
	uint64 nBase;
	
	nBase = nOffset * nBlockSize + psPart->offset * psPart->logical_block_size;

	if( nBlockSize > 1024 )
	{
		return( result ); // AtheOS-FS until now only supports 1k blocks
	}
	
	nError = afs_probe( nDev, nBase, &sFSInfo );
	printf( "afs_probe for dev %ld returned %s (%d)\n", nDev, strerror( nError ), nError );
	if( nError < B_OK )
	{
		return( result );
	}
	strcpy( psPart->volume_name, sFSInfo.volume_name );	
	strcpy( psPart->file_system_short_name, "AtheOS-FS" );
	strcpy( psPart->file_system_long_name, "AtheOS FS" );
	
	result = true;
	return result;
}

void ds_fs_initialize(BMessage *pMessage)
{
	BRect frame;
	BWindow* pWindow;
	InitializeWindow *pInitWindow;
	
	pMessage->FindPointer( "window", ( void** )&pWindow );
	frame = pWindow->Frame();
	frame.left += ( ( frame.Width() - INIT_WINDOW_WIDTH ) / 2 );
	frame.right = frame.left + INIT_WINDOW_WIDTH;
	frame.top += ( ( frame.Height() - INIT_WINDOW_HEIGHT ) / 2 );
	frame.bottom = frame.top + INIT_WINDOW_HEIGHT;
	
	BScreen screen( pWindow );
	BRect screenFrame = screen.Frame();
	screenFrame.InsetBy( 6, 6 );
	if( frame.right > screenFrame.right )
	{
		frame.OffsetBy( screenFrame.right - frame.right, 0 );
	}
	if( frame.bottom > screenFrame.bottom )
	{
		frame.OffsetBy( 0, screenFrame.bottom - frame.bottom );
	}
	if( frame.left < screenFrame.left )
	{
		frame.OffsetBy( screenFrame.left - frame.left, 0 );
	}
	if( frame.top < screenFrame.top )
	{
		frame.OffsetBy( 0, screenFrame.top - frame.top );
	}
	pInitWindow = new InitializeWindow( frame, pMessage );
}
