/*
 *  AFS for BeOS - The native AtheOS file-system for BeOS
 *  Copyright (C) 2001 Ronny Schoebel
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <interface/Box.h>
#include <interface/StringView.h>
#include <interface/Font.h>
#include <app/Message.h>
#include <interface/Alert.h>
#include <interface/Button.h>
#include <interface/TextControl.h>
#include <support/String.h>
#include <support/SupportDefs.h>
#include <storage/Path.h>
#include <storage/FindDirectory.h>
#include <unistd.h>
#include "InitializeWindow.h"

InitializeWindow::InitializeWindow( BRect frame, BMessage *pMessage ) :
	BWindow( frame, "Initialize", B_MODAL_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE )
{
	BFont			font;
	BBox*			pBox;
	BStringView*	pTitleView;
	const char*		pzDeviceName;
	const char*		pzVolumeName;
	char			pzTitle[ 256 ];
	
	if( ( pMessage->FindString( "device", &pzDeviceName ) != B_OK ) ||
		( pMessage->FindInt32( "block_size", &fnBlockSize ) != B_OK ) )
	{
		(new BAlert("Message Data Error", "Sorry, can't initialize:\n\n"
			"Error in Message passed to AtheOS-FS add-on!", "Cancel" ) )->Go();
		Run();
		PostMessage( new BMessage( MSG_CANCEL ) );
		return;
	}
	fDeviceName = pzDeviceName;
	fpMessage = pMessage;
	frame.OffsetTo( 0, 0 );
	frame.InsetBy( -1, -1 );
	frame.top--;				// To make it look like the other add-ons.......
	pBox = new BBox( frame, "root", B_FOLLOW_ALL, B_WILL_DRAW, B_FANCY_BORDER );
	AddChild( pBox );

	sprintf( pzTitle, "%s - %s", TITLE_TEXT, fDeviceName.String() );
	frame.Set( TITLE_X, TITLE_Y, TITLE_X + TITLE_WIDTH, TITLE_Y + TITLE_HEIGHT );
	pTitleView = new BStringView( frame, "title", pzTitle );
	pBox->AddChild( pTitleView );
	font = *be_bold_font;
	font.SetSize( 11.0 );
	pTitleView->SetFont( &font );

	frame.Set(LINE_X, LINE_Y, LINE_X + LINE_WIDTH, LINE_Y + 1);
	pBox->AddChild( new BBox( frame, "line", B_FOLLOW_ALL, B_WILL_DRAW, B_FANCY_BORDER ) );
		
	frame.Set( VOLUME_NAME_X, VOLUME_NAME_Y,
		VOLUME_NAME_X + VOLUME_NAME_WIDTH, VOLUME_NAME_Y + VOLUME_NAME_HEIGHT );
	if( pMessage->FindString( "name", &pzVolumeName ) == B_OK )
	{
		fpVolumeName = new BTextControl( frame, "volume_name", VOLUME_NAME_TEXT,
			pzVolumeName, new BMessage( MSG_VOLUME_NAME ) );
	}
	else
	{
		fpVolumeName = new BTextControl( frame, "volume_name", VOLUME_NAME_TEXT,
			"untitled", new BMessage( MSG_VOLUME_NAME ) );
	}
	fpVolumeName->GetFont( &font );
	fpVolumeName->SetDivider( font.StringWidth( VOLUME_NAME_TEXT ) + 4 );
	( ( BTextView* )fpVolumeName->ChildAt( 0 ) )->SetMaxBytes( 32 );
	pBox->AddChild( fpVolumeName );
	
	frame.Set( BUTTON_OK_X, BUTTON_OK_Y, BUTTON_OK_X + BUTTON_WIDTH, BUTTON_OK_Y + BUTTON_HEIGHT );
	pBox->AddChild( new BButton( frame, "ok_button", BUTTON_OK_TEXT, new BMessage( MSG_OK ) ) );
	
	frame.Set( BUTTON_CANCEL_X, BUTTON_CANCEL_Y, BUTTON_CANCEL_X + BUTTON_WIDTH, BUTTON_CANCEL_Y + BUTTON_HEIGHT );
	pBox->AddChild( new BButton( frame, "cancel_button", BUTTON_CANCEL_TEXT, new BMessage( MSG_CANCEL ) ) );
	
	Show();
}

void InitializeWindow::MessageReceived( BMessage *pMessage )
{
	BWindow*		pWindow;
	bool			bResult = false;
	char			pzDeviceType[ 64 ];
	char			pzAlertText[ 256 ];
	int32			nPartition;
	BAlert*			pAlert;
		
	switch( pMessage->what )
	{
		case MSG_OK:
			
			fpMessage->FindInt32( "partition", &nPartition );
			if( nPartition >= 0 )
			{
				strcpy( pzDeviceType, "partition" );
			}
			else
			{
				strcpy( pzDeviceType, "device" );
			}
			sprintf( pzAlertText, "Initializing will destroy all data on this %s.", pzDeviceType );
			pAlert = new BAlert( "Warning", pzAlertText, "Initialize", "Cancel", NULL,
				B_WIDTH_AS_USUAL, B_WARNING_ALERT );
			pAlert->SetShortcut( 1, B_ESCAPE );
			if( pAlert->Go() == 1 )
			{
				return;
			}
			if( Initialize() != B_OK )
			{
				sprintf( pzAlertText, "Error initializing %s.", pzDeviceType );
				pAlert = new BAlert( "Error", pzAlertText, "Bummer", NULL, NULL,
					B_WIDTH_AS_USUAL, B_STOP_ALERT );
				pAlert->Go();
				return;
			}
			bResult = true;
			
		case MSG_CANCEL:
		
			fpMessage->FindPointer( "window", ( void** )&pWindow );
			fpMessage->AddPointer( "part_window", ( void* )this );
			fpMessage->AddInt32( "part_looper_thread", Thread() );
			fpMessage->AddBool( "result", bResult );
			if( pWindow != NULL )
			{
				pWindow->PostMessage( fpMessage );
			}
			delete fpMessage;
			Quit();
			break;
	
		default:
			
			BWindow::MessageReceived( pMessage );
	}
}

status_t InitializeWindow::Initialize()
{
	BPath		path;
	char 		pzCommand[ B_OS_NAME_LENGTH + B_FILE_NAME_LENGTH + 32 ];
	status_t	nError;
	
	nError = find_directory( B_COMMON_BIN_DIRECTORY, &path );
	if( nError != B_OK )
	{
		return nError;
	}
	sprintf( pzCommand, "mkatheosfs \"%s\" \"%s\"\n", fDeviceName.String(), fpVolumeName->Text() );
	path.Append( pzCommand );
	return( system( path.Path() ) );
}
