/*
 *  AFS - The native AtheOS file-system
 *  Copyright (C) 1999 - 2001 Kurt Skauen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*  Ported to BeOS by Ronny Schoebel, 2001
 *  Original Version: AtheOS 0.3.2
 */


#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <malloc.h>
#include <OS.h>
#include <KernelExport.h>
#include "afsbuild.h"
#include "fsproto.h"
#include "lock.h"
#include "cache.h"
#include "support.h"
#include "macros.h"
#include "afs.h"

#define HT_DEFAULT_SIZE	128
#define HASH(b)			(b)


int afs_init_hash_table( AfsHashTable_s* psTable )
{
	psTable->ht_nSize  = HT_DEFAULT_SIZE;
	psTable->ht_nMask  = psTable->ht_nSize - 1;
	psTable->ht_nCount = 0;

	psTable->ht_apsTable = calloc(1,  psTable->ht_nSize * sizeof(AfsHashEnt_s*));
  
	if ( psTable->ht_apsTable == NULL )
	{
		return( ENOMEM );
	}
	else
	{
		return( B_OK );
	}
}

void afs_delete_hash_table( AfsHashTable_s* psTable )
{
	free( psTable->ht_apsTable );
}


static int afs_resize_hash_table( AfsHashTable_s* psTable )
{
	int	nPrevSize;
	int	nNewSize;
	int   nNewMask;
	int	i;
	int   nHash;
	AfsHashEnt_s** pasNewTable;

	if ( psTable->ht_nSize & psTable->ht_nMask )
	{
		panic("bcache:resize_hash_table() Inconsistency between hashtable size %d and mask %d in bcache!\n",
			psTable->ht_nSize, psTable->ht_nMask );
		return( EINVAL );
	}

	nPrevSize = psTable->ht_nSize;
	nNewSize  = nPrevSize * 2;
	nNewMask  = nNewSize - 1;

//	DPRINTF( "Resize hash table from %d to %d\n", nPrevSize, nNewSize );

	pasNewTable = calloc( 1, nNewSize * sizeof(AfsHashEnt_s*));

	if ( pasNewTable == NULL )
	{
		return( ENOMEM );
	}

	for( i = 0 ; i < nPrevSize ; ++i )
	{
		AfsHashEnt_s*  psEntry;
		AfsHashEnt_s*  psNext;

		for( psEntry = psTable->ht_apsTable[i] ; NULL != psEntry ; psEntry = psNext )
		{
			nHash  = psEntry->he_nHashVal & nNewMask;
			psNext = psEntry->he_psNext;

			psEntry->he_psNext = pasNewTable[ nHash ];
			pasNewTable[ nHash ] = psEntry;
		}
	}
	free( psTable->ht_apsTable );
	psTable->ht_apsTable = pasNewTable;
	psTable->ht_nSize    = nNewSize;
	psTable->ht_nMask    = nNewMask;

	return( 0 );
}

static int afs_hash_insert( AfsHashTable_s* psTable, off_t nBlockNum, AfsHashEnt_s* psNew )
{
	AfsHashEnt_s*	psTmp;
	int				nHash;
	
	nHash = HASH( nBlockNum ) & psTable->ht_nMask;
	
	for( psTmp = psTable->ht_apsTable[nHash] ; psTmp != NULL ; psTmp = psTmp->he_psNext )
	{
		if ( psTmp->he_nBlockNum == nBlockNum )
		{
			break;
		}
	}
	
	if ( psTmp != NULL && psTmp->he_nBlockNum == nBlockNum)
	{
		panic("afs_hash_insert() Entry already in the hash table! Block=%d\n", (int)nBlockNum );
		return( EEXIST );
	}
	
	psNew->he_nBlockNum	= nBlockNum;
	psNew->he_nHashVal	= HASH( nBlockNum );
	
	psNew->he_psNext = psTable->ht_apsTable[ nHash ];
	psTable->ht_apsTable[ nHash ] = psNew;
	
	psTable->ht_nCount++;
	
	if (psTable->ht_nCount >= ((psTable->ht_nSize * 3) / 4))
	{
		afs_resize_hash_table( psTable );
	}
	return( B_OK );
}

static AfsHashEnt_s* afs_hash_lookup( AfsHashTable_s* psTable, off_t nBlockNum )
{
	int				nHash = HASH(nBlockNum) & psTable->ht_nMask;
	AfsHashEnt_s*	psEnt;

	for( psEnt = psTable->ht_apsTable[nHash] ; psEnt != NULL ; psEnt = psEnt->he_psNext )
	{
    	if ( psEnt->he_nBlockNum == nBlockNum )
    	{
			break;
		}
	}
	return( psEnt );
}

static AfsHashEnt_s* afs_hash_delete( AfsHashTable_s* psTable, off_t nBlockNum )
{
	int				nHash	= HASH(nBlockNum) & psTable->ht_nMask;
	AfsHashEnt_s*	psTmp;
	AfsHashEnt_s*	psPrev	= NULL;
	
	for( psTmp = psTable->ht_apsTable[ nHash ] ; NULL != psTmp ; psPrev = psTmp, psTmp = psTmp->he_psNext )
	{
		if ( psTmp->he_nBlockNum == nBlockNum )
		{
			break;
		}
	}
	
	if (psTmp == NULL)
	{
		panic("afs_hash_delete() Entry %d does not exist!\n", (int)nBlockNum );
		return( NULL );
	}
	
	if ( psTable->ht_apsTable[nHash] == psTmp )
	{
		psTable->ht_apsTable[nHash] = psTmp->he_psNext;
	}
	else
	{
		if ( NULL != psPrev )
		{
			psPrev->he_psNext = psTmp->he_psNext;
		}
		else
		{
			panic( "afs_hash_delete() invalid hash table!\n" );
		}
	}
	psTable->ht_nCount--;
	
	if ( psTable->ht_nCount < 0 )
	{
		panic( "afs_hash_delete() ht_nCount got negative value %d!!!!!\n", psTable->ht_nCount );
	}
	return( psTmp );
}

static void afs_add_jb_entry( AfsTransaction_s* psTrans, AfsHashEnt_s* psEntry )
{
	if ( psEntry->he_psTransaction != NULL )
	{
		panic( "afs_add_jb_entry() entry already belongs to a transaction\n" );
		return;
	}
	
	if ( psTrans->at_psFirstEntry != NULL )
	{
		psTrans->at_psFirstEntry->he_psPrevInTrans = psEntry;
	}
	if ( psTrans->at_psLastEntry == NULL )
	{
		psTrans->at_psLastEntry = psEntry;
	}
	psEntry->he_psNextInTrans = psTrans->at_psFirstEntry;
	psEntry->he_psPrevInTrans = NULL;
	psTrans->at_psFirstEntry = psEntry;
	psEntry->he_psTransaction = psTrans;
}

static void afs_remove_jb_entry( AfsTransaction_s* psTrans, AfsHashEnt_s* psEntry )
{
	if ( psEntry->he_psNextInTrans != NULL )
	{
		psEntry->he_psNextInTrans->he_psPrevInTrans = psEntry->he_psPrevInTrans;
	}
	if ( psEntry->he_psPrevInTrans != NULL )
	{
		psEntry->he_psPrevInTrans->he_psNextInTrans = psEntry->he_psNextInTrans;
	}
	if ( psTrans->at_psFirstEntry == psEntry )
	{
		psTrans->at_psFirstEntry = psEntry->he_psNextInTrans;
	}
	if ( psTrans->at_psLastEntry == psEntry )
	{
		psTrans->at_psLastEntry = psEntry->he_psPrevInTrans;
	}
	psEntry->he_psNextInTrans = NULL;
	psEntry->he_psPrevInTrans = NULL;
	psEntry->he_psTransaction = NULL;
}

static int afs_write_superblock( AfsVolume_s* psVolume )
{
    int nError = write_pos( psVolume->av_nDevice, psVolume->av_nSuperBlockPos, psVolume->av_psSuperBlock, AFS_SUPERBLOCK_SIZE );
    if ( nError < 0 )
    {
		DPRINTF( "Error: afs_write_superblock() failed to write superblock : %d\n", nError );
		return( nError );
    }
    else if ( nError != AFS_SUPERBLOCK_SIZE )
    {
		DPRINTF( "Error: afs_write_superblock() failed to write a full superblock : %d\n", nError );
		return( EIO );
    }
    else
    {
		return( B_OK );
    }
}

static void afs_block_flushed_callback( off_t nBlockNum, size_t nNumBlocks, void* pArg )
{
	AfsTransaction_s* psTrans = pArg;
  
	atomic_add( (atomic_t*)&psTrans->at_nPendingLogBlocks, -1 );
	if ( psTrans->at_nPendingLogBlocks < 0 )
	{
		panic( "afs_block_flushed_callback() at_nPendingLogBlocks == %d\n", psTrans->at_nPendingLogBlocks );
	}
}

static AfsTransBuffer_s* afs_alloc_transaction_buffer( AfsVolume_s* psVolume, AfsTransaction_s* psTrans )
{
    AfsTransBuffer_s*	psBuffer;
    uint32				nSize = sizeof( AfsTransBuffer_s ) + 129 * psVolume->av_psSuperBlock->as_nBlockSize;
    area_id				hArea;
    
    nSize = (nSize + B_PAGE_SIZE - 1) & PAGE_MASK;

    psBuffer = NULL;
    hArea = create_area( "afs_trans_buf", (void**)&psBuffer, B_ANY_KERNEL_ADDRESS, nSize, B_FULL_LOCK, B_WRITE_AREA | B_READ_AREA);

    if ( hArea < 0 )
    {
		DPRINTF( "afs_alloc_transaction_buffer() failed to create trasaction buffer\n" );
		return( NULL );
    }

    psBuffer->atb_pBuffer = (off_t*) (psBuffer+1);
    memset( psBuffer->atb_pBuffer, 0, psVolume->av_psSuperBlock->as_nBlockSize ); // Clear the index block
    psBuffer->atb_hAreaID = hArea;
    psBuffer->atb_nMaxBlockCount = 128;
    psBuffer->atb_nBlockCount = 0;
    psBuffer->atb_psNext = psTrans->at_psFirstBuffer;
    psTrans->at_psFirstBuffer = psBuffer;
    return( psBuffer );
}

static void afs_free_transaction_buffer( AfsVolume_s* psVolume, AfsTransBuffer_s* psBuffer )
{
	delete_area( psBuffer->atb_hAreaID );
}

int afs_replay_log( AfsVolume_s* psVolume )
{
	AfsSuperBlock_s*	psSuperBlock = psVolume->av_psSuperBlock;
	const int			nBlockSize = psSuperBlock->as_nBlockSize;
	off_t				nLogBlock;
	int					nError;
	off_t*				pIndex;
	void*				pBlockBuffer;
	
	if ( psSuperBlock->as_nValidLogBlocks == 0 )
	{
		psSuperBlock->as_nLogStart = psVolume->av_nLoggStart;
		return( B_OK );
	}
	DPRINTF( "afs_replay_log() %d log blocks from %d\n",
		psSuperBlock->as_nValidLogBlocks, (int)psSuperBlock->as_nLogStart );
	
	pIndex = malloc( nBlockSize );
	if ( pIndex == NULL )
	{
		DPRINTF( "Error: afs_replay_log() no memory for index block\n" );
		return( ENOMEM );
	}
	pBlockBuffer = malloc( nBlockSize );
	if ( pBlockBuffer == NULL )
	{
		DPRINTF( "Error: afs_replay_log() no memory for block buffer\n" );
		free( pIndex );
		return( ENOMEM );
	}
	
	nLogBlock = psSuperBlock->as_nLogStart;
	
	while( psSuperBlock->as_nValidLogBlocks )
	{
		int i;
		nError = read_phys_blocks( psVolume->av_nDevice, nLogBlock, pIndex, 1, nBlockSize );
		if ( nError < 0 )
		{
			DPRINTF( "Error: afs_replay_log() failed to read index block\n" );
			goto error;
		}
		nLogBlock++;
		psSuperBlock->as_nValidLogBlocks--;
		if ( nLogBlock > psVolume->av_nLoggEnd )
		{
			nLogBlock = psVolume->av_nLoggStart;
		}
		for ( i = 0 ; i < 128 ; i++ )
		{
			if ( pIndex[i] == 0 )
			{
				break;
			}
			DPRINTF( "afs_replay_log() write back block %d to %d\n", (int) nLogBlock, (int) pIndex[i] );
	
			if ( pIndex[i] >= psSuperBlock->as_nNumBlocks )
			{
				DPRINTF( "Panic : attempt to write back block belonging outside the disk!!\n" );
				goto error;
			}
			nError = read_phys_blocks( psVolume->av_nDevice, nLogBlock++, pBlockBuffer, 1, nBlockSize );
			psSuperBlock->as_nValidLogBlocks--;
			if ( nError < 0 )
			{
				DPRINTF( "Panic: afs_replay_log() failed to read journal block\n" );
				goto error;
			}
			nError = write_phys_blocks( psVolume->av_nDevice, pIndex[i], pBlockBuffer, 1, nBlockSize );
			if ( nError < 0 )
			{
				DPRINTF( "Panic: afs_replay_log() failed to write back journal block\n" );
				goto error;
			}
			if ( nLogBlock > psVolume->av_nLoggEnd )
			{
				nLogBlock = psVolume->av_nLoggStart;
			}
		}
	}
	psSuperBlock->as_nLogStart = psVolume->av_nLoggStart;
	psSuperBlock->as_nValidLogBlocks = 0;
	
	afs_write_superblock( psVolume );
	//  write_phys_blocks( psVolume->av_nDevice, 0, psVolume->av_psSuperBlock, 1, nBlockSize );
	
	free( pBlockBuffer );
	free( pIndex );
	return( B_OK );
error:
	free( pBlockBuffer );
	free( pIndex );
	return( nError );
}

static int afs_create_transaction( AfsVolume_s* psVolume )
{
	AfsTransaction_s* psPrevTrans = psVolume->av_psFirstTrans;
	AfsTransaction_s* psTrans;
	
	kassertw( psPrevTrans == NULL || psPrevTrans->at_bWrittenToLog );
	
	psTrans = calloc( 1, sizeof( AfsTransaction_s ));
	if ( psTrans == NULL )
	{
		DPRINTF( "Error: afs_create_transaction() no memory for transaction header\n" );
		return( ENOMEM );
	}
	// Calculate the starting point in the journal of the new transaction
	if ( psPrevTrans != NULL )
	{
		off_t nPrevEnd = psPrevTrans->at_nLogStart + psPrevTrans->at_nBlockCount;
	
		if ( nPrevEnd <= psVolume->av_nLoggEnd )
		{
			psTrans->at_nLogStart = nPrevEnd;
		}
		else
		{
			psTrans->at_nLogStart = psVolume->av_nLoggStart + (nPrevEnd - psVolume->av_nLoggEnd) - 1;
		}
	}
	else
	{
		psTrans->at_nLogStart = psVolume->av_nLoggStart;
	}
	psTrans->at_nBlockCount = 0;
	
//	DPRINTF( "afs_create_transaction() new transaction start at %d\n", (int) psTrans->at_nLogStart );
	
	// Link the new transaction into the volumes list.
	if ( psVolume->av_psLastTrans == NULL )
	{
		psVolume->av_psLastTrans = psTrans;
	}
	if ( psVolume->av_psFirstTrans != NULL )
	{
		psVolume->av_psFirstTrans->at_psPrev = psTrans;
	}
	psTrans->at_psNext = psVolume->av_psFirstTrans;
	psTrans->at_psPrev = NULL;
	psVolume->av_psFirstTrans = psTrans;
	return( B_OK );
}

int afs_begin_transaction( AfsVolume_s* psVolume )
{
	int nError = B_OK;
	SEM_LOCK( psVolume->av_hJournalLock );

//	DPRINTF( "Begin transaction\n" );
	if ( psVolume->av_nTransactionStarted != 0 )
	{
    	panic( "attempt to nest calls to afs_begin_transaction()\n" );
		return( EINVAL );
	}
	if ( psVolume->as_nTransAllocatedBlocks != 0 )
	{
		panic( "afs_begin_transaction() %Ld blocks allocated without a transaction\n", psVolume->as_nTransAllocatedBlocks );
		return( EINVAL );
	}
	psVolume->av_nTransactionStarted++;

	SEM_LOCK( psVolume->av_hJournalListsLock );
	if ( psVolume->av_psFirstTrans == NULL || psVolume->av_psFirstTrans->at_bWrittenToLog )
	{
		nError = afs_create_transaction( psVolume );
		if( nError < B_OK )
		{
			psVolume->av_nTransactionStarted--;
		}
	} 
	SEM_UNLOCK( psVolume->av_hJournalListsLock );
	return( nError );
}

static int afs_write_transaction_to_disk( AfsVolume_s* psVolume, AfsTransaction_s* psTrans )
{
	const int nBlockSize = psVolume->av_psSuperBlock->as_nBlockSize;
	AfsTransBuffer_s* psBuf;
	AfsHashEnt_s*     psEntry;
	int               nIndex = 0;

	for ( psBuf = psTrans->at_psFirstBuffer ; psBuf != NULL ; psBuf = psBuf->atb_psNext )
	{
		int i;

		for ( i = 0 ; i < psBuf->atb_nBlockCount ; ++i )
		{
			off_t			nBlock = psBuf->atb_pBuffer[i];
			void*			pLogBuffer = ((uint8*)psBuf->atb_pBuffer) + (i+1) * nBlockSize;
			int				nError;

			if ( nIndex++ < psTrans->at_nWrittenBlocks )
			{
				DPRINTF( "afs_write_transaction_to_disk() block %d already written. Skipping\n", nIndex );
				continue;
			}
			nError = write_logged_blocks( psVolume->av_nDevice, nBlock, pLogBuffer, 1, nBlockSize,
				afs_block_flushed_callback, psTrans );
			if ( nError < 0 )
			{
				// FIXME: We should try a few times and then mark the FS read-only and give up.
				DPRINTF( "Error: afs_write_transaction_to_disk() failed to write block (err=%d) will try again later\n", nError );
				return( nError );
			}
/*			psEntry = afs_hash_lookup( &psVolume->av_sLogHashTable, nBlock );
			if ( psEntry != NULL && psEntry->he_pBuffer == pLogBuffer )
			{
				if ( afs_hash_delete( &psVolume->av_sLogHashTable, nBlock ) != psEntry )
				{
					panic( "afs_write_transaction() failed to delete entry from hash-table\n" );
				}
				free( psEntry );
			}*/
			atomic_add( (atomic_t*)&psTrans->at_nPendingLogBlocks, 1 );
			atomic_add( (atomic_t*)&psTrans->at_nWrittenBlocks, 1 );
		}
	}
	nIndex = 0;
	while ( psTrans->at_psFirstEntry != NULL )
	{
		psEntry = psTrans->at_psFirstEntry;
		psTrans->at_psFirstEntry = psEntry->he_psNextInTrans;
		if ( psEntry->he_psTransBuf == NULL )
		{ 
//			DPRINTF( "afs_write_transaction_to_disk() remove block %Ld from hash table (%p)\n", psEntry->he_nBlockNum, psTrans );
		
			if ( afs_hash_delete( &psVolume->av_sLogHashTable, psEntry->he_nBlockNum ) != psEntry )
			{
				panic( "afs_write_transaction_to_disk() failed to delete entry from hash-table\n" );
			}
			free( psEntry );
		}
		else
		{
//			printk( "afs_write_transaction_to_disk() block %Ld to current. Not from hash table (%p)\n", psEntry->he_nBlockNum, psTrans );
			psEntry->he_psTransaction = NULL;
		}
		nIndex++;
	}
//	if ( nIndex > 0 )
//	{
//		DPRINTF( "afs_write_transaction_to_disk() removed %d hash-table entries\n", nIndex );
//	}
	psTrans->at_psLastEntry = NULL;
	psTrans->at_bWrittenToDisk = true;
	return( B_OK );
}

static void afs_force_flush_transaction( AfsVolume_s* psVolume, AfsTransaction_s* psTrans )
{
	AfsTransBuffer_s* psBuf;

	for ( psBuf = psTrans->at_psFirstBuffer ; psBuf != NULL ; psBuf = psBuf->atb_psNext )
	{
		int i;
		for ( i = 0 ; i < psBuf->atb_nBlockCount ; ++i )
		{
	    	flush_blocks( psVolume->av_nDevice, psBuf->atb_pBuffer[i], 1 );
		}
	}
}


static int afs_discard_flushed_journal_entries( AfsVolume_s* psVolume )
{
	AfsTransaction_s* psTrans;
	
	psTrans = psVolume->av_psLastTrans;
	if ( psTrans == NULL )
	{
		return( B_OK );
	}
	if ( psTrans->at_bWrittenToLog == false || psTrans->at_bWrittenToDisk == false )
	{
		return( B_OK );
	}
	if ( psTrans->at_nPendingLogBlocks > 0 )
	{
		return( B_OK );
	}
	// Unlink the finnished transaction.
	psVolume->av_psLastTrans = psTrans->at_psPrev;
	if ( psTrans->at_psPrev != NULL )
	{
		psTrans->at_psPrev->at_psNext = NULL;
	}
	else
	{
		psVolume->av_psFirstTrans = NULL;
	}
	psVolume->av_nPendingLogBlocks -= psTrans->at_nBlockCount;
	kassertw( psVolume->av_nPendingLogBlocks >= 0 );
	
	// Update the super-block.
	
	psVolume->av_psSuperBlock->as_nValidLogBlocks -= psTrans->at_nBlockCount;
	kassertw( psVolume->av_psSuperBlock->as_nValidLogBlocks >= 0 );
	
	if ( psVolume->av_psLastTrans == NULL )
	{
		if ( psVolume->av_psSuperBlock->as_nValidLogBlocks != 0 )
		{
			panic( "afs_discard_flushed_journal_entries(): No more transactions but %d blocks are pending\n",
				psVolume->av_psSuperBlock->as_nValidLogBlocks);
			return( EINVAL );
		}
		psVolume->av_psSuperBlock->as_nLogStart = psVolume->av_nLoggStart;
	}
	else
	{
		kassertw( psVolume->av_psLastTrans->at_bWrittenToDisk == false );
		psVolume->av_psSuperBlock->as_nLogStart = psVolume->av_psLastTrans->at_nLogStart;
	}
//	DPRINTF( "afs_discard_flushed_journal_entries() New journal range = %d (%d)\n",
//		(int)psVolume->av_psSuperBlock->as_nLogStart, (int)psVolume->av_psSuperBlock->as_nValidLogBlocks );
	
	while( psTrans->at_psFirstBuffer != NULL )
	{
		AfsTransBuffer_s* psBuf = psTrans->at_psFirstBuffer;
		psTrans->at_psFirstBuffer = psBuf->atb_psNext;
		afs_free_transaction_buffer( psVolume, psBuf );
	}
	free( psTrans );
	
	afs_write_superblock( psVolume );
	
	// If the now last transaction is written to log we write it to the disk now.
	psTrans = psVolume->av_psLastTrans;
	if ( psTrans != NULL && psTrans->at_bWrittenToLog )
	{
		afs_write_transaction_to_disk( psVolume, psTrans );
	}
	return( B_OK );
}

static int afs_wait_for_log_space( AfsVolume_s* psVolume, int nBlockCount )
{
	int nLoops = 0;
	
	while( psVolume->av_nPendingLogBlocks + nBlockCount > psVolume->av_nJournalSize )
	{
		if ( nLoops++ > 500 )
		{
			DPRINTF( "Wait for %d log blocks to be free (%d)\n", nBlockCount, psVolume->av_nPendingLogBlocks );
		}
		if ( psVolume->av_psLastTrans != NULL )
		{
			afs_force_flush_transaction( psVolume, psVolume->av_psLastTrans );
		}
//		flush_device_cache( psVolume->av_nDevice, true );
		if ( psVolume->av_psLastTrans != NULL &&
			psVolume->av_psLastTrans->at_bWrittenToLog &&
			psVolume->av_psLastTrans->at_bWrittenToDisk == false)
		{
			afs_write_transaction_to_disk( psVolume, psVolume->av_psLastTrans );
		}
		afs_discard_flushed_journal_entries( psVolume );
	}
	return( B_OK );
}

static int afs_write_transaction_to_log( AfsVolume_s* psVolume )
{
	const int			nBlockSize = psVolume->av_psSuperBlock->as_nBlockSize;
	AfsTransaction_s*	psTrans = psVolume->av_psFirstTrans;
	AfsTransBuffer_s*	psBuf;
	off_t				nBlockAddr;
	int					nError = 0;
	int					nBlocksToWrite = psVolume->av_nOldBlockCount;
	int					nTotalBlockCount = 0;
	
	if ( psTrans == NULL || psTrans->at_bWrittenToLog == true )
	{
		return( B_OK );
	}

	if ( psTrans->at_nLogStart < psVolume->av_nLoggStart )
    {
		panic( "afs_write_transaction_to_log() First block start before log-area. Log: %Ld->%Ld, Log-start: %Ld\n",
			psVolume->av_nLoggStart, psVolume->av_nLoggEnd, psTrans->at_nLogStart );
	}
	if ( psTrans->at_nLogStart > psVolume->av_nLoggEnd )
	{
		panic( "afs_write_transaction_to_log() First block start after log-area. Log: %Ld->%Ld, Log-start: %Ld\n",
			psVolume->av_nLoggStart, psVolume->av_nLoggEnd, psTrans->at_nLogStart );
	}
	nBlockAddr = psTrans->at_nLogStart;
//	DPRINTF( "Write transaction from %d (%d blocks)\n", (int)psTrans->at_nLogStart, psTrans->at_nBlockCount );
	for ( psBuf = psTrans->at_psFirstBuffer ; psBuf != NULL ; psBuf = psBuf->atb_psNext )
	{
		int nBlockCount = psBuf->atb_nBlockCount + 1; // Data blocks + index block;
	
		if ( nBlockAddr + nBlockCount <= psVolume->av_nLoggEnd + 1 )
		{
			nError = write_phys_blocks( psVolume->av_nDevice, nBlockAddr, psBuf->atb_pBuffer,
				nBlockCount, nBlockSize );
//			DPRINTF( "Write one run %d-%d\n", (int)nBlockAddr, (int)(nBlockAddr + nBlockCount - 1) );
			if ( nError < 0 )
			{
				DPRINTF( "Error: afs_write_transaction_to_log() failed to write unsplitted run\n" );
				DPRINTF( "Error: %Ld -> %Ld (%d) Error = %d\n", nBlockAddr, nBlockAddr + nBlockCount*nBlockSize - 1, nBlockSize, nError );
				goto error;
			}
			nBlockAddr += nBlockCount;
		}
		else
		{
			int nFirstSize = psVolume->av_nLoggEnd - nBlockAddr + 1;
			int nSecondSize = nBlockCount - nFirstSize;
	    
			if ( nFirstSize <= 0 || nSecondSize <= 0 )
			{
				panic( "afs_write_transaction_to_log() got negative block-run. Log: %Ld->%Ld, Trans: %Ld,%d, Sizes: %d,%d\n",
					psVolume->av_nLoggStart, psVolume->av_nLoggEnd, psTrans->at_nLogStart, psTrans->at_nBlockCount, nFirstSize, nSecondSize );
			}
			
			nError = write_phys_blocks( psVolume->av_nDevice, nBlockAddr, psBuf->atb_pBuffer,
				nFirstSize, nBlockSize );
			if ( nError < 0 )
			{
				DPRINTF( "Error: afs_write_transaction_to_log() failed to write first run of splitted write\n" );
				DPRINTF( "Error: %Ld -> %Ld (%d) Error = %d\n", nBlockAddr, nBlockAddr + nFirstSize*nBlockSize - 1, nBlockSize, nError );
				goto error;
			}
			nBlockAddr = psVolume->av_nLoggStart;
			nError = write_phys_blocks( psVolume->av_nDevice, nBlockAddr,
				( ( uint8* )psBuf->atb_pBuffer ) + ( nFirstSize * nBlockSize ),
				nSecondSize, nBlockSize );
			if ( nError < B_OK )
			{
				DPRINTF( "Error: afs_write_transaction_to_log() failed to write second run of splitted write\n" );
				DPRINTF( "Error: %Ld -> %Ld (%d) Error = %d\n", psVolume->av_nLoggStart,
					psVolume->av_nLoggStart + (nBlockCount-nFirstSize)*nBlockSize - 1, nBlockSize, nError );
				goto error;
			}
			nBlockAddr += nSecondSize;
		}
		nTotalBlockCount += nBlockCount;
	}
	psVolume->av_nOldBlockCount -= nTotalBlockCount;
	psTrans->at_bWrittenToLog = true;
	
	if ( psVolume->av_nOldBlockCount != 0 )
	{
		panic( "afs_write_transaction_to_log() mismatching block count! av_nOldBlockCount = %d\n",
			psVolume->av_nOldBlockCount );
	}
	
	// Update the range in super-block
	psVolume->av_psSuperBlock->as_nValidLogBlocks += nBlocksToWrite;
//	DPRINTF( "afs_write_transaction_to_log() New journal range = %d (%d)\n",
//		(int)psVolume->av_psSuperBlock->as_nLogStart, psVolume->av_psSuperBlock->as_nValidLogBlocks );
	
	afs_write_superblock( psVolume );
	
	if ( psVolume->av_psLastTrans != NULL &&
		psVolume->av_psLastTrans->at_bWrittenToLog &&
		psVolume->av_psLastTrans->at_bWrittenToDisk == false)
	{
		afs_write_transaction_to_disk( psVolume, psVolume->av_psLastTrans );
	}
	return( B_OK );
//	return( afs_create_transaction( psVolume ) );
error:
	return( nError );
}

static void afs_validate_transaction( AfsVolume_s* psVolume, AfsTransaction_s* psTrans )
{
	AfsTransBuffer_s* psBuf;
	int nBlockCount = 0;
	
	for ( psBuf = psTrans->at_psFirstBuffer ; psBuf != NULL ; psBuf = psBuf->atb_psNext )
	{
		int i;
		nBlockCount += psBuf->atb_nBlockCount + 1;
		for ( i = 0 ; i < psBuf->atb_nBlockCount ; ++i )
		{
			if ( psBuf->atb_pBuffer[i] == 0 )
			{
				panic( "afs_validate_transaction() transaction buffer have 0 reference at index %d\n", i );
			}
		}
		for ( ; i < 128 ; ++i )
		{
			if ( psBuf->atb_pBuffer[i] != 0 )
			{
				panic( "afs_validate_transaction() transaction buffer have non-0 reference at index %d\n", i );
			}
		}
	}	
	if ( nBlockCount != psTrans->at_nBlockCount )
	{
		panic( "afs_validate_transaction() mismatch between at_nBlockCount (%d) and actual block count (%d)\n",
			psTrans->at_nBlockCount, nBlockCount );
	}
}

static int afs_merge_transactions( AfsVolume_s* psVolume )
{
	const int			nBlockSize = psVolume->av_psSuperBlock->as_nBlockSize;
	AfsTransaction_s*	psTrans = psVolume->av_psFirstTrans;
	AfsJournalBlock_s*	psBlock;
	int					nNewBlocks = 0;
	int					nError = 0;
	int					nNew = 0;
	int					nMerged = 0;
	
	if ( psTrans == NULL )
	{
		panic( "afs_merge_transactions() someting whent terribly wrong! psTrans ended up as a NULL pointer\n" );
		return( EINVAL );
	}
	if ( psTrans->at_bWrittenToLog )
	{
		DPRINTF( "afs_merge_transactions() the transaction is already written to the log\n" );
		return( EINVAL );
	}
	afs_wait_for_log_space( psVolume, psVolume->av_nNewBlockCount + psVolume->av_nJournalSize / 128 );
	
	// Here we merge the blocks touched during the current transaction with blocks
	// from previous transactions.
    for ( psBlock = psVolume->av_psFirstLogBlock ; psBlock != NULL ; psBlock = psBlock->jb_psNext )
	{
		AfsTransBuffer_s*	psBuf;
		AfsHashEnt_s*		psEntry;
	
		psEntry = afs_hash_lookup( &psVolume->av_sLogHashTable, psBlock->jb_nBlock );
		if ( psEntry == NULL )
		{
			panic( "afs_merge_transactions() can't find block %Ld in hash-table\n", psBlock->jb_nBlock );
			return( EINVAL );
		}
		if ( psEntry->he_psTransBuf != psBlock )
		{
			panic( "afs_merge_transactions() hash-table entry does not point to most resent block!\n" );
			return( EINVAL );
		}
		if ( psEntry->he_psTransaction != psTrans || psEntry->he_pBuffer == NULL )
		{
			// If new or not our's we must alloc a new block
			if ( psTrans->at_psFirstBuffer == NULL || psTrans->at_psFirstBuffer->atb_nBlockCount == psTrans->at_psFirstBuffer->atb_nMaxBlockCount )
			{
				psBuf = afs_alloc_transaction_buffer( psVolume, psTrans );
				if ( psBuf == NULL )
				{
					DPRINTF( "Error: afs_merge_transactions() no memory for transaction buffer\n" );
					nError = ENOMEM;
					goto error;
				}
				psTrans->at_nBlockCount++; // Need space for the index block
				psVolume->av_nOldBlockCount++;
				psVolume->av_nPendingLogBlocks++;
			}
			else
			{
				psBuf = psTrans->at_psFirstBuffer;
			}
			// Think twice before re-ordering the next few lines (then leave it as it is).
			if ( psBlock->jb_nBlock == 0 )
			{
				panic( "afs_merge_transactions() block refere to disk-address 0!\n" );
			}
			psBuf->atb_pBuffer[psBuf->atb_nBlockCount] = psBlock->jb_nBlock;
			psBuf->atb_nBlockCount++;
			psBlock->jb_pNewBuffer = ((uint8*)psBuf->atb_pBuffer) + (psBuf->atb_nBlockCount)*nBlockSize;
			nNewBlocks++;
		}
	}

	while( psVolume->av_psFirstLogBlock != NULL )
	{
		AfsHashEnt_s* psEntry;
	
		psBlock = psVolume->av_psFirstLogBlock;
		psVolume->av_psFirstLogBlock = psBlock->jb_psNext;
	
		psEntry = afs_hash_lookup( &psVolume->av_sLogHashTable, psBlock->jb_nBlock );
		if ( psEntry == NULL )
		{
			panic( "afs_merge_transactions:2() can't find block in hash-table\n" );
			return( EINVAL );
		}
		if ( psEntry->he_psTransBuf != psBlock )
		{
			panic( "afs_merge_transactions() hash-table entry does not point to most resent block!\n" );
			return( EINVAL );
		}
		if ( psEntry->he_psTransaction != psTrans || psEntry->he_pBuffer == NULL )
		{
			nNew++;
			memcpy( psBlock->jb_pNewBuffer, psBlock->jb_pData, nBlockSize );
			psEntry->he_pBuffer = psBlock->jb_pNewBuffer;
			if ( psEntry->he_psTransaction != psTrans )
			{
				if( psEntry->he_psTransaction != NULL )
				{
					afs_remove_jb_entry( psEntry->he_psTransaction, psEntry );
				}
				afs_add_jb_entry( psTrans, psEntry );
			}
		}
		else
		{
			nMerged++;
			memcpy( psEntry->he_pBuffer, psBlock->jb_pData, nBlockSize );
			psVolume->av_nNewBlockCount--;
		}
		psEntry->he_psTransBuf = NULL;
		free( psBlock );
	}
//	DPRINTF( "Merge transaction %d new and %d merged\n", nNew, nMerged );
	
	psTrans->at_nBlockCount			+= nNewBlocks;
	psVolume->av_nOldBlockCount		+= nNewBlocks;
	psVolume->av_nPendingLogBlocks	+= nNewBlocks;
	psVolume->av_nNewBlockCount		-= nNewBlocks;
	if ( psVolume->av_nNewBlockCount != 0 )
	{
		panic( "afs_merge_transactions() mismatching block count! av_nNewBlockCount = %d\n",
			psVolume->av_nNewBlockCount );
	}
	afs_validate_transaction( psVolume, psTrans );
	return( B_OK );

error:
	
	while ( nNewBlocks > 0 )
	{
		if ( psTrans->at_psFirstBuffer == NULL )
		{
			panic( "afs_merge_transactions() to few transaction buffers. %d blocks left\n", nNewBlocks );
		}
		if ( psTrans->at_psFirstBuffer->atb_nBlockCount <= nNewBlocks )
		{
			AfsTransBuffer_s* psNext = psTrans->at_psFirstBuffer->atb_psNext;
		
			nNewBlocks -= psTrans->at_psFirstBuffer->atb_nBlockCount;
	
			DPRINTF( "afs_merge_transactions() delete buffer with %d blocks\n", psTrans->at_psFirstBuffer->atb_nBlockCount );
			afs_free_transaction_buffer( psVolume, psTrans->at_psFirstBuffer );
			psTrans->at_psFirstBuffer = psNext;
	
			psTrans->at_nBlockCount--;
			psVolume->av_nOldBlockCount--;
			psVolume->av_nPendingLogBlocks--;
		}
		else
		{
			DPRINTF( "afs_merge_transactions() strip %d blocks from buffer\n", nNewBlocks );
			psTrans->at_psFirstBuffer->atb_nBlockCount -= nNewBlocks;
			memset( psTrans->at_psFirstBuffer->atb_pBuffer + psTrans->at_psFirstBuffer->atb_nBlockCount, 0, nNewBlocks * sizeof(off_t) );
			nNewBlocks = 0;
		}
	}
	afs_validate_transaction( psVolume, psTrans );
	return( nError );
}

static int afs_discard_transactions( AfsVolume_s* psVolume )
{
	int nNumReverted  = 0;
	int nNumDiscarded = 0;
	int i;
	
	while( psVolume->av_psFirstLogBlock != NULL )
	{
		AfsJournalBlock_s* psBlock = psVolume->av_psFirstLogBlock;
		AfsHashEnt_s*      psEntry;
	
		psEntry = afs_hash_lookup( &psVolume->av_sLogHashTable, psBlock->jb_nBlock );
		if ( psEntry == NULL )
		{
			panic( "afs_discard_transactions() can't find block in hash-table\n" );
			return( EINVAL );
		}
		if ( psEntry->he_psTransBuf != psBlock )
		{
			panic( "afs_discard_transactions() hash-table entry does not point to most resent block!\n" );
			return( EINVAL );
		}
		if ( psEntry->he_pBuffer == NULL )
		{
			// It don't belong to any older transactions either so we nukes it.
//			DPRINTF( "afs_discard_transactions() remove block %Ld from hash table (%p)\n", psBlock->jb_nBlock, psEntry->he_psTransaction );
			if ( afs_hash_delete( &psVolume->av_sLogHashTable, psBlock->jb_nBlock ) != psEntry )
			{
				panic( "afs_discard_transactions() failed to delete entry from hash-table\n" );
			}
			if( psEntry->he_psTransaction != NULL )
			{
				afs_remove_jb_entry( psEntry->he_psTransaction, psEntry );
			}
			free( psEntry );
			nNumDiscarded++;
		}
		else
		{
			nNumReverted++;
		}
		psEntry->he_psTransBuf = NULL;
	
		psVolume->av_psFirstLogBlock = psBlock->jb_psNext;
		free( psBlock );
		psVolume->av_nNewBlockCount--;
	}
	if ( psVolume->av_nNewBlockCount != 0 )
	{
		panic( "afs_discard_transactions() mismatching block count! av_nNewBlockCount = %d\n",
			psVolume->av_nNewBlockCount );
	}
//	DPRINTF( "afs_discard_transactions() reverted %d blocks and discarded %d blocks. Reverted %Ld allocated blocks\n",
//		nNumReverted, nNumDiscarded, psVolume->as_nTransAllocatedBlocks );
	psVolume->av_psSuperBlock->as_nUsedBlocks -= psVolume->as_nTransAllocatedBlocks;
	psVolume->as_nTransAllocatedBlocks = 0;
	
	for ( i = 0 ; i < psVolume->av_sLogHashTable.ht_nCount ; ++i )
	{
		AfsHashEnt_s* psEntry;
		
		for ( psEntry = psVolume->av_sLogHashTable.ht_apsTable[i] ; psEntry != NULL ; psEntry = psEntry->he_psNext )
		{
			if ( psEntry->he_psTransBuf != NULL )
			{
				panic( "afs_discard_transactions() hashtable entry still points to block header\n" );
			}
			if ( psEntry->he_pBuffer == NULL )
			{
				panic( "afs_discard_transactions() hashtable entry have no buffer\n" );
			}
		}
	}
	return( B_OK );
}

int afs_wait_for_block_to_leave_log( AfsVolume_s* psVolume, off_t nBlock )
{
	int nRun = 0;
	int nError = B_OK;
	
	while( afs_hash_lookup( &psVolume->av_sLogHashTable, nBlock ) != NULL )
	{
		DPRINTF( "Wait for block %Ld to leave the log\n", nBlock );
		if ( nRun == 2 )
		{
			nError = afs_write_transaction_to_log( psVolume );
			if( nError >= B_OK )
			{
				nError = afs_create_transaction( psVolume );
			}
		}
		if ( psVolume->av_psLastTrans != NULL )
		{
			afs_force_flush_transaction( psVolume, psVolume->av_psLastTrans );
		}
//		flush_device_cache( psVolume->av_nDevice, true );
		if ( psVolume->av_psLastTrans != NULL &&
			psVolume->av_psLastTrans->at_bWrittenToLog &&
			psVolume->av_psLastTrans->at_bWrittenToDisk == false)
		{
			afs_write_transaction_to_disk( psVolume, psVolume->av_psLastTrans );
		}
		afs_discard_flushed_journal_entries( psVolume );
		nRun++;
//		if ( nRun > 500 )
//		{
//			DPRINTF( "Panic: Block %d won't leave the log\n", (int)nBlock );
//			break;
//		}
	}
	return( nError );
}

int afs_checkpoint_journal( AfsVolume_s* psVolume )
{
	status_t nError = B_OK;
//	DPRINTF( "afs_checkpoint_journal()\n" );
	
	SEM_LOCK( psVolume->av_hJournalListsLock );
	if ( psVolume->av_nOldBlockCount + psVolume->av_nNewBlockCount + psVolume->av_nJournalSize / 128 > psVolume->av_nJournalSize )
	{
		nError = afs_write_transaction_to_log( psVolume );
		if ( nError >= B_OK )
		{
			nError = afs_create_transaction( psVolume );
		}
	}
	if ( nError >= 0 )
	{
//		DPRINTF( "afs_create_transaction() merge transaction\n" );
		nError = afs_merge_transactions( psVolume );
		if ( nError >= 0 )
		{
			afs_discard_flushed_journal_entries( psVolume );
		}
	}
	if ( nError >= 0 )
	{
		psVolume->as_nTransAllocatedBlocks = 0;
	}
	SEM_UNLOCK( psVolume->av_hJournalListsLock );
	return( nError );
}

int afs_end_transaction( AfsVolume_s* psVolume, bool bSubmitChanges )
{
	status_t	nError = B_OK;
//	off_t		nUsedBlocks;
	
	SEM_LOCK( psVolume->av_hJournalListsLock );
	if ( psVolume->av_nOldBlockCount + psVolume->av_nNewBlockCount > 128 )
	{
		nError = afs_write_transaction_to_log( psVolume );
		if( nError >= B_OK )
		{
			nError = afs_create_transaction( psVolume );
		}
	}
	if( nError >= B_OK && bSubmitChanges )
	{
		nError = afs_merge_transactions( psVolume );
		if ( nError < B_OK )
		{
			DPRINTF( "Error: afs_end_transaction() failed to merge transaction: %ld\n", nError );
			afs_discard_transactions( psVolume );
		}
		else
		{
		    psVolume->as_nTransAllocatedBlocks = 0;
		}
	}
	else
	{
		afs_discard_transactions( psVolume );
	}	
	afs_discard_flushed_journal_entries( psVolume );
	SEM_UNLOCK( psVolume->av_hJournalListsLock );
	
	psVolume->av_nTransactionStarted--;
	if ( psVolume->av_nTransactionStarted < 0 )
	{
		panic( "unbalanced call to afs_end_transaction()\n" );
	}
	psVolume->av_nLastJournalAccesTime = system_time();
/*
	nUsedBlocks = afs_count_used_blocks( psVolume );

	if ( psVolume->av_psSuperBlock->as_nUsedBlocks != nUsedBlocks )
	{
		DPRINTF( "Error: afs_end_transaction() Inconsistency between super-block and allocation bitmap!\n" );
		DPRINTF( "Superblock: %Ld used blocks\n", psVolume->av_psSuperBlock->as_nUsedBlocks );
		DPRINTF( "Bitmap:     %Ld used blocks\n", nUsedBlocks );
//		psVolume->av_psSuperBlock->as_nUsedBlocks = nUsedBlocks;
	}
*/
	SEM_UNLOCK( psVolume->av_hJournalLock );
	return( nError );
}

static uint32 afs_journal_flusher( void* pData )
{
	AfsVolume_s* volatile psVolume = pData;

	while( psVolume->av_bRunJournalFlusher )
	{
		snooze( 1000000 );

		SEM_LOCK( psVolume->av_hJournalLock );
		SEM_LOCK( psVolume->av_hJournalListsLock );

//		if ( psVolume->av_nOldBlockCount + psVolume->av_nNewBlockCount > 128 )
//		{
//			afs_write_transaction_to_log( psVolume );
//		}
		if ( psVolume->av_nNewBlockCount != 0 )
		{
			panic( "afs_journal_flusher() found a half-finished transaction in the journal!\n" );
			continue;
//			afs_merge_transactions( psVolume );
		}
//		if ( psVolume->av_nOldBlockCount != 0 )
//		{
			afs_write_transaction_to_log( psVolume );
//		}
		afs_discard_flushed_journal_entries( psVolume );
		SEM_UNLOCK( psVolume->av_hJournalListsLock );
		SEM_UNLOCK( psVolume->av_hJournalLock );
	}
	psVolume->av_bRunJournalFlusher = true; // Handshake with unmount() to tell that we are about to leave.
	return( B_OK );
}

int afs_start_journal_flusher( AfsVolume_s* psVolume )
{
	thread_id hThread;
  
	hThread = spawn_kernel_thread((thread_entry)afs_journal_flusher, "afs_log_flusher", B_NORMAL_PRIORITY, psVolume );
	if ( hThread >= 0 )
	{
    	resume_thread( hThread );
  	}
	return( hThread );
}

void afs_flush_journal( AfsVolume_s* psVolume )
{
	DPRINTF( "Flushing journal:\n" );

	SEM_LOCK( psVolume->av_hJournalLock );

	flush_device( psVolume->av_nDevice, false );

	if ( psVolume->av_nOldBlockCount + psVolume->av_nNewBlockCount + psVolume->av_nJournalSize / 128 > psVolume->av_nJournalSize )
	{
		DPRINTF( "flush old journal entry\n" );
		afs_write_transaction_to_log( psVolume );
	}
	DPRINTF( "Merge last transaction\n" );
	if ( psVolume->av_nNewBlockCount != 0 )
	{
		panic( "afs_flush_journal() found a half-finished transaction in the journal!\n" );
//	afs_merge_transactions( psVolume );
	}
	DPRINTF( "Write last transaction to the journal\n" );
	afs_write_transaction_to_log( psVolume );
	DPRINTF( "Wait for the transactions to finnish\n" );
	afs_wait_for_log_space( psVolume, psVolume->av_nJournalSize );
	DPRINTF( "All journal entries flushed\n" );
	
	SEM_UNLOCK( psVolume->av_hJournalLock );
}

int afs_logged_read( AfsVolume_s* psVolume, void* pBuffer, off_t nBlock )
{
	const int		nBlockSize = psVolume->av_psSuperBlock->as_nBlockSize;
	AfsHashEnt_s*	psEntry;
	int				nError;
	
	psVolume->av_nLastJournalAccesTime = system_time();
	
	SEM_LOCK( psVolume->av_hJournalListsLock );
	
	psEntry = afs_hash_lookup( &psVolume->av_sLogHashTable, nBlock );
	
	if ( psEntry != NULL )
	{
//		DPRINTF( "afs_logged_read() found block %Ld in log\n", nBlock );
		if ( psEntry->he_psTransBuf != NULL )
		{
			memcpy( pBuffer, psEntry->he_psTransBuf->jb_pData, nBlockSize );
			nError = B_OK;
		}
		else if ( psEntry->he_pBuffer != NULL )
		{
			memcpy( pBuffer, psEntry->he_pBuffer, nBlockSize );
			nError = B_OK;
		}
		else
		{
			panic( "afs_logged_read() both buffer pointers of hash entry is NULL!!\n" );
			nError = EINVAL;
		}
	}
	else
	{
//		DPRINTF( "afs_logged_read() load block %Ld\n", nBlock );
		nError = cached_read( psVolume->av_nDevice, nBlock, pBuffer, 1, nBlockSize );
	}
	SEM_UNLOCK( psVolume->av_hJournalListsLock );
	return( nError );
}

int afs_logged_write( AfsVolume_s* psVolume, const void* pBuffer, off_t nBlock )
{
	const int			nBlockSize = psVolume->av_psSuperBlock->as_nBlockSize;
	AfsJournalBlock_s*	psBlock;
	AfsHashEnt_s*		psEntry;
	int					nError;
	
	SEM_LOCK( psVolume->av_hJournalListsLock );
	
	if ( psVolume->av_nTransactionStarted != 1 )
	{
		panic( "afs_logged_write() called while av_nTransactionStarted = %d\n", psVolume->av_nTransactionStarted );
		nError = EINVAL;
		goto error;
	}
	
	psEntry = afs_hash_lookup( &psVolume->av_sLogHashTable, nBlock );
	
	if ( psEntry != NULL && psEntry->he_psTransBuf != NULL )
	{
		memcpy( psEntry->he_psTransBuf->jb_pData, pBuffer, nBlockSize );
		goto done;
	}
	
	psBlock = calloc( 1, sizeof( AfsJournalBlock_s ) + nBlockSize );
	if ( psBlock == NULL )
	{
		DPRINTF( "Error: afs_logged_write() no memory for logg block\n" );
		nError = ENOMEM;
		goto error;
	}
	
	if ( psEntry == NULL )
	{
		psEntry = calloc( 1, sizeof( AfsHashEnt_s ) );
		if ( psEntry == NULL )
		{
			DPRINTF( "Error: afs_logged_write() no memory for hash-table entry\n" );
			nError = ENOMEM;
			free( psBlock );
			goto error;
		}
//		DPRINTF( "afs_logged_write() add block %Ld to hash table (%p)\n", nBlock, psVolume->av_psFirstTrans );
		afs_hash_insert( &psVolume->av_sLogHashTable, nBlock, psEntry );
		afs_add_jb_entry( psVolume->av_psFirstTrans, psEntry );
	}
/*
	else
	{
		if ( psEntry->he_psTransaction == NULL )
		{
			panic( "afs_logged_write() found entry in hash-table not belonging to any transaction\n" );
		}
	}
*/	
	psVolume->av_nNewBlockCount++;
	psBlock->jb_pData = (uint8*) (psBlock+1);
	psBlock->jb_nBlock = nBlock;
	
	psBlock->jb_psNext = psVolume->av_psFirstLogBlock;
	psVolume->av_psFirstLogBlock = psBlock;
	memcpy( psBlock->jb_pData, pBuffer, nBlockSize );
	
	psEntry->he_psTransBuf = psBlock;
	
done:
	psVolume->av_nLastJournalAccesTime = system_time();
	SEM_UNLOCK( psVolume->av_hJournalListsLock );
	return( B_OK );

error:
	psVolume->av_nLastJournalAccesTime = system_time();
	SEM_UNLOCK( psVolume->av_hJournalListsLock );
	return( nError );
}
