/*
 *  AFS for BeOS - The native AtheOS file-system for BeOS
 *  Copyright (C) 2001 Ronny Schoebel
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <SupportDefs.h>
#include <malloc.h>
#include <string.h>
#include "afsbuild.h"
#include "macros.h"
#include "afs.h"

int64 afs_get_device_size( int nDev )
{
	device_geometry		sGeo;
	partition_info		sPart;
//	size_t nSize;
	int nError = B_OK;

	nError = ioctl( nDev, B_GET_PARTITION_INFO, &sPart );
	if(nError == B_OK)
	{
		return( sPart.size );
	}
	nError = ioctl( nDev, B_GET_GEOMETRY, &sGeo );
	if ( nError == B_OK )
	{
		return( sGeo.cylinder_count * sGeo.sectors_per_track * 
			sGeo.head_count * sGeo.bytes_per_sector );
	}
	else
	{
		struct stat	   sStat;
    
    	nError = fstat( nDev, &sStat );
		if ( nError < 0 )
		{
			return( nError );
		}
		return( sStat.st_size );
	}
}

static int afs_initialize( const char* pzDevPath,
	const char *pzVolName, void* pArgs, size_t nArgLen )
{
	AfsSuperBlock_s*	psSuperBlock;
	int					nBitmapBlkCnt;
	uint32*				panBitmap;
	AfsInode_s*			psInode;
	int					i;
	
	int					nBlockSize = 1024;
	int64				nNumBlocks = 1024 * 1024;
	int					nLogSize   = 2048;
	int					nBootLoaderSize = 32;
	int					nDev;
	off_t				nBitmapPos;
	off_t				nCurPos;
	
	nDev = open( pzDevPath, O_RDWR );
	
	if ( nDev < 0 )
	{
		DPRINTF( "Error: afs_initialize() failed to open device %s\n", pzDevPath );
		return( nDev );
	}
	
	nNumBlocks = afs_get_device_size( nDev );
	
	if ( nNumBlocks < 0 )
	{
		DPRINTF( "afs_initialize() failed to get device size : %Ld\n", nNumBlocks );
		close( nDev );
		return( nNumBlocks );
	}
	
	nNumBlocks /= nBlockSize;
	
	if ( nNumBlocks < 3000 )
	{
		DPRINTF( "Error: afs_initialize() Disk to small, cant initialize\n" );
		close( nDev );
		return( EINVAL );
	}
	
	kassertw( 1024 == nBlockSize || 2048 == nBlockSize || 4096 == nBlockSize || 8192 == nBlockSize );
	
	
	nBitmapBlkCnt = (nNumBlocks + nBlockSize * 8 - 1) / (nBlockSize * 8);
	
	psSuperBlock = malloc( nBlockSize );
	
	if ( NULL == psSuperBlock )
	{
		close( nDev );
		return( ENOMEM );
	}
	
	psInode = malloc( nBlockSize );
	
	if ( NULL == psInode )
	{
		free( psSuperBlock );
		close( nDev );
		return( ENOMEM );
	}
	
	panBitmap = malloc( nBlockSize * nBitmapBlkCnt );
	
	if ( NULL == panBitmap )
	{
		free( psSuperBlock );
		free( psInode );
		close( nDev );
		return( ENOMEM );
	}
	
	/* +--------------------------+
	* |       Boot block         |
	* |         0-1023           |
	* +--------------------------+
	* |       Super block        |
	* |       1024 - 2047        |
	* +~~~~~~~~~~~~~~~~~~~~~~~~~~+
	* |    Boot loader area      |
	* +~~~~~~~~~~~~~~~~~~~~~~~~~~+
	* |      Bitmap blocks       |
	* +--------------------------+
	* |     Journal blocks       |
	* +--------------------------+
	*/
	
	memset( psSuperBlock, 0, nBlockSize );
	memset( psInode, 0, nBlockSize );
	memset( panBitmap, 0, nBlockSize * nBitmapBlkCnt );
	
	strncpy( psSuperBlock->as_zName, pzVolName, 31 );
	psSuperBlock->as_zName[31]		= '\0';
	
	psSuperBlock->as_nMagic1	= SUPER_BLOCK_MAGIC1;
	psSuperBlock->as_nByteOrder	= BO_LITTLE_ENDIAN;
	psSuperBlock->as_nBlockSize	= nBlockSize;
	
	for ( psSuperBlock->as_nBlockShift = 0 ;
		(1 << psSuperBlock->as_nBlockShift) < nBlockSize ;
		++psSuperBlock->as_nBlockShift )
	/*** EMPTY ***/;
	kassertw( (1 << psSuperBlock->as_nBlockShift) == nBlockSize );
	
	
	nBitmapPos = (nBootLoaderSize * nBlockSize + AFS_SUPERBLOCK_SIZE + 1024 + nBlockSize - 1) / nBlockSize;
	nCurPos = nBitmapPos + nBitmapBlkCnt;
	
	psSuperBlock->as_nNumBlocks	 = nNumBlocks;
	psSuperBlock->as_nUsedBlocks = 1 + nBitmapBlkCnt + 1;	// super block + bitmap + root dir
	psSuperBlock->as_nBootLoaderSize = nBootLoaderSize;
	psSuperBlock->as_nInodeSize	= nBlockSize;
	
	psSuperBlock->as_nMagic2	    = SUPER_BLOCK_MAGIC2;
	psSuperBlock->as_nBlockPerGroup = nBlockSize * 8;	// Number of blocks per allocation group (Max 65536)
	// Number of bits to shift a group number to get a byte address.
	
	for ( psSuperBlock->as_nAllocGrpShift = 0 ;
		(1 << psSuperBlock->as_nAllocGrpShift) < psSuperBlock->as_nBlockPerGroup * nBlockSize ;
		psSuperBlock->as_nAllocGrpShift++ )
	{
	/*** EMPTY ***/;
	}
	
	psSuperBlock->as_nAllocGroupCount = (nNumBlocks + psSuperBlock->as_nBlockPerGroup - 1)  / psSuperBlock->as_nBlockPerGroup;
	
	psSuperBlock->as_nFlags	= 0;
	
	psSuperBlock->as_sLogBlock.group = 0;
	psSuperBlock->as_sLogBlock.start = nCurPos;
	psSuperBlock->as_sLogBlock.len   = nLogSize;
	
	psSuperBlock->as_nLogStart	     = nCurPos;
	psSuperBlock->as_nValidLogBlocks = 0;
	psSuperBlock->as_nLogSize	     = nLogSize;
	nCurPos += nLogSize;
	
	psSuperBlock->as_nMagic3 = SUPER_BLOCK_MAGIC3;
	
	psSuperBlock->as_nUsedBlocks = nCurPos + 3;
	
	if ( psSuperBlock->as_nNumBlocks < psSuperBlock->as_nBlockPerGroup * 8 + 4 )
	{
		psSuperBlock->as_sRootDir.group = 0;
		psSuperBlock->as_sRootDir.start = nCurPos++;
		psSuperBlock->as_sRootDir.len   = 1;
	
		psSuperBlock->as_sIndexDir.group = 0;
		psSuperBlock->as_sIndexDir.start = nCurPos++;
		psSuperBlock->as_sIndexDir.len   = 1;
	
		psSuperBlock->as_sDeletedFiles.group = 0;
		psSuperBlock->as_sDeletedFiles.start = nCurPos++;
		psSuperBlock->as_sDeletedFiles.len   = 1;
	}
	else
	{
		psSuperBlock->as_sRootDir.group = 8;
		psSuperBlock->as_sRootDir.start = 0;
		psSuperBlock->as_sRootDir.len   = 1;
	
		psSuperBlock->as_sIndexDir.group = 8;
		psSuperBlock->as_sIndexDir.start = 1;
		psSuperBlock->as_sIndexDir.len   = 1;
	
		psSuperBlock->as_sDeletedFiles.group = 8;
		psSuperBlock->as_sDeletedFiles.start = 2;
		psSuperBlock->as_sDeletedFiles.len   = 1;
	}
	
	
	DPRINTF( "NumBlocks         = %Ld\n", psSuperBlock->as_nNumBlocks );
	DPRINTF( "Block shift       = %ld\n", psSuperBlock->as_nBlockShift );
	DPRINTF( "Blocks per group  = %ld\n", psSuperBlock->as_nBlockPerGroup );
	DPRINTF( "Alloc group shift = %ld\n", psSuperBlock->as_nAllocGrpShift );
	DPRINTF( "Alloc group count = %ld\n", psSuperBlock->as_nAllocGroupCount );
	DPRINTF( "Bitmap block count= %d\n",  nBitmapBlkCnt );
	
	DPRINTF( "Prealloc super block, bitmap and root inode\n" );
	
	// Allocate super block, boot-loader area, bitmap blocks and the log.
	for ( i = 0 ; i < nCurPos ; ++i )
	{
		panBitmap[ i / 32 ] |= 1 << (i % 32);
	}
	// Prealloc blocks in the last bitmap block that resides outside the disk
	for ( i = nNumBlocks ; i < nBitmapBlkCnt * nBlockSize * 8 ; ++i )
	{
		panBitmap[ i / 32 ] |= 1 << (i % 32);
	}
	
	i = afs_run_to_num( psSuperBlock, &psSuperBlock->as_sRootDir );
	panBitmap[ i / 32 ] |= 1 << (i % 32);	// Allocate root dir inode.
	
	i = afs_run_to_num( psSuperBlock, &psSuperBlock->as_sIndexDir );
	panBitmap[ i / 32 ] |= 1 << (i % 32);	// Allocate index dir inode.
	
	i = afs_run_to_num( psSuperBlock, &psSuperBlock->as_sDeletedFiles );
	panBitmap[ i / 32 ] |= 1 << (i % 32); // Allocate delete-me dir inode.
	
	
	DPRINTF( "Write super block\n" );
	
	write_pos( nDev, 1024, psSuperBlock, AFS_SUPERBLOCK_SIZE );
	write_pos( nDev, nBitmapPos * nBlockSize, panBitmap, nBitmapBlkCnt * nBlockSize );
	
	psInode->ai_nMagic1		= INODE_MAGIC;
	psInode->ai_nUID		= 0;
	psInode->ai_nGID		= 0;
	psInode->ai_nFlags		= INF_ATTRIBUTES | INF_LOGGED | INF_USED;
	psInode->ai_nMode		= S_IFDIR | S_IRWXUGO;
	psInode->ai_nIndexType	= e_KeyTypeString;
	psInode->ai_nLinkCount	= 1;
	psInode->ai_nCreateTime	= real_time_clock_usecs();
	psInode->ai_nModifiedTime	= psInode->ai_nCreateTime;
	psInode->ai_nInodeSize	= nBlockSize;
	
	psInode->ai_sInodeNum	= psSuperBlock->as_sRootDir;
	write_pos( nDev, afs_run_to_num( psSuperBlock, &psInode->ai_sInodeNum ) * nBlockSize, psInode, nBlockSize );
	
	psInode->ai_sInodeNum	= psSuperBlock->as_sIndexDir;
	write_pos( nDev, afs_run_to_num( psSuperBlock, &psInode->ai_sInodeNum ) * nBlockSize, psInode, nBlockSize );
	
	psInode->ai_sInodeNum	= psSuperBlock->as_sDeletedFiles;
	write_pos( nDev, afs_run_to_num( psSuperBlock, &psInode->ai_sInodeNum ) * nBlockSize, psInode, nBlockSize );
	
	// Clear the boot-block to avoid misdetection of the FS later.
	memset( psSuperBlock, 0, 1024 );
	write_pos( nDev, 0, psSuperBlock, 1024 );
	
	free( psSuperBlock );
	free( panBitmap );
	free( psInode );
	
	DPRINTF( "AFS volume %s formated\n", pzVolName );
	close( nDev );
	
	return( B_OK );
}

void usage( char *pStr )
{
	printf( "Usage: %s <devicename> <volumename>\n" , pStr );
	puts( "Where <devicename> is the path to the device to be initialised" );
	puts( "      <volumename> is the name of the new volume." );
}

int main( int argc, char **argv )
{
	status_t nError = B_OK;
	if( argc < 3 )
	{
		usage( argv[0] );
	}
	else
	{
		nError = afs_initialize( argv[1], argv[2], NULL, 0 );
		if( nError < B_OK )
		{
			printf("Error while initializing volume! (%s)\n", strerror( nError ) );
		}
	}
	return nError;
}
